from typing import Any, Dict, Hashable, Mapping, TypeVar, overload

from typing_aliases import StringDict, StringMapping

__all__ = ("merge",)

Q = TypeVar("Q", bound=Hashable)
T = TypeVar("T")


@overload
def merge(*mappings: Mapping[Q, T]) -> Dict[Q, T]:
    ...


@overload
def merge(*mappings: StringMapping[T], **keywords: T) -> StringDict[T]:
    ...


def merge(*mappings: Mapping[Any, Any], **keywords: Any) -> Dict[Any, Any]:
    """Merges multiple `mappings` and `keywords` into one dictionary.

    Arguments:
        *mappings: Mappings to merge.
        **keywords: Keywords to add to the merged dictionary.

    Returns:
        The merged dictionary.
    """
    merged: Dict[Any, Any] = {}

    for mapping in mappings:
        merged.update(mapping)

    merged.update(keywords)

    return merged



from typing import Iterable, Iterator, TypeVar

from typing_aliases import Binary, Unary
from wraps.option import Option, is_some

__all__ = ("filter_map_option", "scan")

S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")


def scan(state: S, function: Binary[S, T, Option[U]], iterable: Iterable[T]) -> Iterator[U]:
    for item in iterable:
        option = function(state, item)

        if is_some(option):
            yield option.unwrap()

        else:
            break


def filter_map_option(function: Unary[T, Option[U]], iterable: Iterable[T]) -> Iterator[U]:
    for option in map(function, iterable):
        if is_some(option):
            yield option.unwrap()



from __future__ import annotations

from abc import abstractmethod as required
from typing import Protocol, TypeVar, runtime_checkable

__all__ = ("Sum", "Product")

S = TypeVar("S", bound="Sum")


@runtime_checkable
class Sum(Protocol):
    """Represents types for which adding `self: S` to `other: S` returns `S`."""

    @required
    def __add__(self: S, __other: S) -> S:
        raise NotImplementedError


P = TypeVar("P", bound="Product")


@runtime_checkable
class Product(Protocol):
    """Represents types for which multiplying `self: P` with `other: P` returns `P`."""

    @required
    def __mul__(self: P, __other: P) -> P:
        raise NotImplementedError



from __future__ import annotations

from typing import Any, TypeVar, Union

from solus import Singleton
from typing_extensions import TypeGuard
from wraps.option import Null, Option, Some

__all__ = (
    # markers
    "Marker",
    "marker",
    # no default
    "NoDefault",
    "no_default",
    # wrap marked
    "wrap_marked",
    # type guards
    "is_marker",
    "is_not_marker",
    "is_no_default",
    "is_not_no_default",
)

T = TypeVar("T")


class NoDefault(Singleton):
    """Represents the absence of default values."""


no_default = NoDefault()
"""The instance of [`NoDefault`][iters.types.NoDefault]."""


NoDefaultOr = Union[NoDefault, T]


def is_no_default(item: Any) -> TypeGuard[NoDefault]:
    """Checks if the `item` is [`NoDefault`][iters.types.NoDefault].

    Returns:
        Whether the `item` is [`NoDefault`][iters.types.NoDefault].
    """
    return item is no_default


def is_not_no_default(item: NoDefaultOr[T]) -> TypeGuard[T]:
    """Checks if the `item` is not [`NoDefault`][iters.types.NoDefault].

    Returns:
        Whether the `item` is not [`NoDefault`][iters.types.NoDefault].
    """
    return item is not no_default


class Marker(Singleton):
    """Represents markers used for various checks."""


marker = Marker()
"""The instance of [`Marker`][iters.types.Marker]."""


MarkerOr = Union[Marker, T]


def is_marker(item: Any) -> TypeGuard[Marker]:
    """Checks if the `item` is [`Marker`][iters.types.Marker].

    Returns:
        Whether the `item` is [`Marker`][iters.types.Marker].
    """
    return item is marker


def is_not_marker(item: MarkerOr[T]) -> TypeGuard[T]:
    """Checks if the `item` is not [`Marker`][iters.types.Marker].

    Returns:
        Whether the `item` is not [`Marker`][iters.types.Marker].
    """
    return item is not marker


def wrap_marked(item: MarkerOr[T]) -> Option[T]:
    return Some(item) if is_not_marker(item) else Null()



from __future__ import annotations

from typing import Sequence, TypeVar, Union, final, overload

from attrs import frozen
from typing_aliases import is_slice
from wraps.wraps import wrap_option

__all__ = ("SequenceView", "sequence_view")

T = TypeVar("T")
U = TypeVar("U")


@final
@frozen()
class SequenceView(Sequence[T]):
    """Represents views over sequences."""

    sequence: Sequence[T]
    """The sequence to view into."""

    @classmethod
    def create(cls, sequence: Sequence[U]) -> SequenceView[U]:
        return cls(sequence)  # type: ignore

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> SequenceView[T]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[T, SequenceView[T]]:
        if is_slice(index):
            return self.create(self.sequence[index])

        return self.sequence[index]  # type: ignore

    def __len__(self) -> int:
        return len(self.sequence)

    @wrap_option
    def get(self, index: int) -> T:
        return self.sequence[index]


sequence_view = SequenceView
